
///////////////////////////////////////////////////////////////////////////
// This program is free software: you can redistribute it and/or modify  //
// it under the terms of the version 3 of the GNU General Public License //
// as published by the Free Software Foundation.                         //
//                                                                       //
// This program is distributed in the hope that it will be useful, but   //
// WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      //
// General Public License for more details.                              //
//                                                                       //
// You should have received a copy of the GNU General Public License     //
// along with this program. If not, see <http://www.gnu.org/licenses/>.  //
//                                                                       //
// Written by Francois Fleuret                                           //
// (C) Idiap Research Institute                                          //
//                                                                       //
// Contact <francois.fleuret@idiap.ch> for comments & bug reports        //
///////////////////////////////////////////////////////////////////////////

#include "pose_cell_hierarchy.h"
#include "gaussian.h"

PoseCellHierarchy::PoseCellHierarchy() {
  _belly_cells = 0;
}

PoseCellHierarchy::PoseCellHierarchy(LabelledImagePool *train_pool) {
  _nb_levels = global.nb_levels;
  _min_head_radius = global.min_head_radius;
  _max_head_radius = global.max_head_radius;
  _root_cell_nb_xy_per_radius = global.root_cell_nb_xy_per_radius;

  LabelledImage *image;
  int nb_total_targets = 0;
  for(int i = 0; i < train_pool->nb_images(); i++) {
    image = train_pool->grab_image(i);
    // We are going to symmetrize
    nb_total_targets += 2 * image->nb_targets();
    train_pool->release_image(i);
  }

  RelativeBellyPoseCell targets[nb_total_targets];

  int u = 0;
  for(int i = 0; i < train_pool->nb_images(); i++) {
    image = train_pool->grab_image(i);

    PoseCellSet cell_set;
    add_root_cells(image, &cell_set);

    for(int t = 0; t < image->nb_targets(); t++) {
      Pose pose = *image->get_target_pose(t);
      Pose coarse;

      cell_set.get_containing_cell(&pose)->get_centroid(&coarse);

      targets[u]._belly_xc.set((pose._belly_xc - coarse._head_xc) / coarse._head_radius);
      targets[u]._belly_yc.set((pose._belly_yc - coarse._head_yc) / coarse._head_radius);
      u++;

      pose.horizontal_flip(image->width());

      cell_set.get_containing_cell(&pose)->get_centroid(&coarse);

      targets[u]._belly_xc.set((pose._belly_xc - coarse._head_xc) / coarse._head_radius);
      targets[u]._belly_yc.set((pose._belly_yc - coarse._head_yc) / coarse._head_radius);
      u++;
    }

    train_pool->release_image(i);
  }

  scalar_t fattening = 1.1;

  Interval belly_rxc, belly_ryc;

  belly_rxc.set(&targets[0]._belly_xc);
  belly_ryc.set(&targets[0]._belly_yc);

  for(int t = 0; t < nb_total_targets; t++) {
    belly_rxc.swallow(&targets[t]._belly_xc);
    belly_ryc.swallow(&targets[t]._belly_yc);
  }

  belly_rxc.min *= fattening;
  belly_rxc.max *= fattening;
  belly_ryc.min *= fattening;
  belly_ryc.max *= fattening;

  scalar_t belly_rxc_min = belly_resolution * floor(belly_rxc.min / belly_resolution);
  int nb_belly_rxc = int(ceil((belly_rxc.max - belly_rxc_min) / belly_resolution));

  scalar_t belly_ryc_min = belly_resolution * floor(belly_ryc.min / belly_resolution);
  int nb_belly_ryc = int(ceil((belly_ryc.max - belly_ryc_min) / belly_resolution));

  int used[nb_belly_rxc * nb_belly_rxc];

  for(int k = 0; k < nb_belly_rxc * nb_belly_ryc; k++) {
    used[k] = 1;
  }

  // An ugly way to compute the convexe enveloppe

  for(scalar_t alpha = 0; alpha < M_PI * 2; alpha += (2 * M_PI) / 100) {
    scalar_t vx = cos(alpha), vy = sin(alpha);
    scalar_t rho = 0;

    for(int t = 0; t < nb_total_targets; t++) {
      rho = min(rho, vx * targets[t]._belly_xc.middle() + vy * targets[t]._belly_yc.middle());
    }

    rho *= fattening;

    for(int j = 0; j < nb_belly_ryc; j++) {
      for(int i = 0; i < nb_belly_rxc; i++) {
        if(
           vx * (scalar_t(i + 0) * belly_resolution + belly_rxc_min) +
           vy * (scalar_t(j + 0) * belly_resolution + belly_ryc_min) < rho
           &&
           vx * (scalar_t(i + 1) * belly_resolution + belly_rxc_min) +
           vy * (scalar_t(j + 0) * belly_resolution + belly_ryc_min) < rho
           &&
           vx * (scalar_t(i + 0) * belly_resolution + belly_rxc_min) +
           vy * (scalar_t(j + 1) * belly_resolution + belly_ryc_min) < rho
           &&
           vx * (scalar_t(i + 1) * belly_resolution + belly_rxc_min) +
           vy * (scalar_t(j + 1) * belly_resolution + belly_ryc_min) < rho
           ) {
          used[i + j * nb_belly_rxc] = 0;
        }
      }
    }
  }

  _nb_belly_cells = 0;
  for(int j = 0; j < nb_belly_ryc; j++) {
    for(int i = 0; i < nb_belly_rxc; i++) {
      if(used[i + nb_belly_rxc * j]) {
        _nb_belly_cells++;
      }
    }
  }

  _belly_cells = new RelativeBellyPoseCell[_nb_belly_cells];

  int k = 0;
  for(int j = 0; j < nb_belly_ryc; j++) {
    for(int i = 0; i < nb_belly_rxc; i++) {

      if(used[i + nb_belly_rxc * j]) {

        RelativeBellyPoseCell mother;

        scalar_t x = scalar_t(i) * belly_resolution + belly_rxc_min;
        scalar_t y = scalar_t(j) * belly_resolution + belly_ryc_min;

        mother._belly_xc.set(x, x + belly_resolution);
        mother._belly_yc.set(y, y + belly_resolution);

        _belly_cells[k++] = mother;
      }
    }
  }
}

PoseCellHierarchy::~PoseCellHierarchy() {
  delete[] _belly_cells;
}

int PoseCellHierarchy::nb_levels() {
  return _nb_levels;
}

void PoseCellHierarchy::get_containing_cell(Image *image, int level,
                                            Pose *pose, PoseCell *result_cell) {
  PoseCellSet cell_set;

  for(int l = 0; l < level + 1; l++) {
    cell_set.erase_content();
    if(l == 0) {
      add_root_cells(image, &cell_set);
    } else {
      add_subcells(l, result_cell, &cell_set);
    }

    *result_cell = *(cell_set.get_containing_cell(pose));
  }
}

void PoseCellHierarchy::add_root_cells(Image *image, PoseCellSet *cell_set) {

  const int nb_scales = int((log(_max_head_radius) - log(_min_head_radius)) / log(2) *
                            global.nb_scales_per_power_of_two);

  scalar_t alpha = log(_min_head_radius);
  scalar_t beta = log(2) / scalar_t(global.nb_scales_per_power_of_two);

  for(int s = 0; s < nb_scales; s++) {
    scalar_t cell_xy_size = exp(alpha + scalar_t(s) * beta) / global.root_cell_nb_xy_per_radius;
    PoseCell cell;
    cell._head_radius.min = exp(alpha + scalar_t(s) * beta);
    cell._head_radius.max = exp(alpha + scalar_t(s+1) * beta);
    cell._head_tilt.min = -M_PI;
    cell._head_tilt.max = M_PI;
    for(scalar_t y = 0; y < image->height(); y += cell_xy_size)
      for(scalar_t x = 0; x < image->width(); x += cell_xy_size) {
        cell._head_xc.min = x;
        cell._head_xc.max = x + cell_xy_size;
        cell._head_yc.min = y;
        cell._head_yc.max = y + cell_xy_size;
        cell._belly_xc.min = cell._head_xc.min - pseudo_infty;
        cell._belly_xc.max = cell._head_xc.max + pseudo_infty;
        cell._belly_yc.min = cell._head_yc.min - pseudo_infty;
        cell._belly_yc.max = cell._head_yc.max + pseudo_infty;
        cell_set->add_cell(&cell);
      }
  }
}

void PoseCellHierarchy::add_subcells(int level, PoseCell *root,
                                     PoseCellSet *cell_set) {

  switch(level) {

  case 1:
    {
      // Here we split the belly-center coordinate cell part
      PoseCell cell = *root;
      scalar_t r = sqrt(cell._head_radius.min * cell._head_radius.max);
      scalar_t x = (cell._head_xc.min + cell._head_xc.max) / 2.0;
      scalar_t y = (cell._head_yc.min + cell._head_yc.max) / 2.0;
      for(int k = 0; k < _nb_belly_cells; k++) {
        cell._belly_xc.min = (_belly_cells[k]._belly_xc.min * r) + x;
        cell._belly_xc.max = (_belly_cells[k]._belly_xc.max * r) + x;
        cell._belly_yc.min = (_belly_cells[k]._belly_yc.min * r) + y;
        cell._belly_yc.max = (_belly_cells[k]._belly_yc.max * r) + y;
        cell_set->add_cell(&cell);
      }
    }
    break;

  default:
    {
      cerr << "Inconsistent level in PoseCellHierarchy::add_subcells" << endl;
      abort();
    }
    break;
  }
}


int PoseCellHierarchy::nb_incompatible_poses(LabelledImagePool *pool) {
  PoseCell target_cell;
  PoseCellSet cell_set;
  LabelledImage *image;

  int nb_errors = 0;

  for(int i = 0; i < pool->nb_images(); i++) {
    image = pool->grab_image(i);

    for(int t = 0; t < image->nb_targets(); t++) {
      cell_set.erase_content();

      int error_level = -1;

      for(int l = 0; error_level < 0 && l < _nb_levels; l++) {
        cell_set.erase_content();

        if(l == 0) {
          add_root_cells(image, &cell_set);
        } else {
          add_subcells(l, &target_cell, &cell_set);
        }

        int nb_compliant = 0;

        for(int c = 0; c < cell_set.nb_cells(); c++) {
          if(cell_set.get_cell(c)->contains(image->get_target_pose(t))) {
            target_cell = *(cell_set.get_cell(c));
            nb_compliant++;
          }
        }

        if(nb_compliant != 1) {
          error_level = l;
        }
      }

      if(error_level >= 0) {
        nb_errors++;
      }
    }

    pool->release_image(i);
  }

  return nb_errors;
}

void PoseCellHierarchy::write(ostream *os) {
  write_var(os, &_min_head_radius);
  write_var(os, &_max_head_radius);
  write_var(os, &_root_cell_nb_xy_per_radius);
  write_var(os, &_nb_belly_cells);
  for(int k = 0; k < _nb_belly_cells; k++)
    write_var(os, &_belly_cells[k]);
}

void PoseCellHierarchy::read(istream *is) {
  delete[] _belly_cells;
  read_var(is, &_min_head_radius);
  read_var(is, &_max_head_radius);
  read_var(is, &_root_cell_nb_xy_per_radius);
  read_var(is, &_nb_belly_cells);
  delete[] _belly_cells;
  _belly_cells = new RelativeBellyPoseCell[_nb_belly_cells];
  for(int k = 0; k < _nb_belly_cells; k++) {
    read_var(is, &_belly_cells[k]);
  }
}
